#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <vfw.h>  // to stop attract video
#include "launcher.h"
#include "drawing.h"
#include "mame.h"
#include "mameparser.h"
#include <mmsystem.h>	// for sndPlaySound()
#include "displayMenu.h"
#include "ini.h"
#include "scrolltext.h"
#include "input.h"
#include "menu.h"
#include "message.h"


#ifdef DEBUG
#include "guicon.h"
#endif

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow);
int runMessageLoop();

// Run once at startup
BOOL initialize();

// Run once at shutdown
BOOL deinitialize();

// Hide the mouse pointer
void RemoveCursor();

// Show the mouse pointer
void DisplayCursor();

// Process any key strokes that have been made
void ProcessKeyStrokes(HWND hwnd);

#if DEBUG
void printWinVersion() {
	OSVERSIONINFO osv;
	osv.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&osv);
	printf("Windows version %d.%d.%d %s\n", osv.dwMajorVersion, osv.dwMinorVersion, osv.dwBuildNumber & 0xffff, osv.szCSDVersion);

/*
Windows 95 retail, OEM		4.00.950
Windows 95 retail SP1		4.00.950A
OEM Service Release 1		4.00.950A
OEM Service Release 2		4.00.1111* (4.00.950B)
OEM Service Release 2.1		4.03.1212-1214* (4.00.950B)
OEM Service Release 2.5		4.03.1214* (4.00.950C)
Windows 98 retail, OEM		4.10.1998
Windows 98 Second Edition	4.10.2222A
Windows ME 					4.10.3000
Windows 2000				5.00.2195
*/
}
#endif

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
	WNDCLASS wc;
	int retval;

#if DEBUG
	// Create a console and output to it in debug mode
	RedirectIOToNewConsole();

	printf("%s\n", VERSION);
	printWinVersion();
	printf("\n");
#endif

	SetConsoleTitle(VERSION);

	if (strlen(lpCmdLine) > 0)
		iniFilename = lpCmdLine;
	else
		iniFilename = DEFAULTINIFILENAME;

#if DEBUG
	printf("Ini file: [%s]\n", iniFilename);
#endif

	hInst=hInstance;
	hwndMain=NULL;

	wc.style = 0;
	wc.lpfnWndProc = (WNDPROC)WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInst;
	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(BLACK_BRUSH);
	wc.lpszMenuName =  "MainMenu";
	wc.lpszClassName = "MainWndClass";
	if(!RegisterClass(&wc)) {
		MessageBox(NULL, "Failed to register window class!", "Error", MB_OK);
		return FALSE;
	}

	hwndMain = CreateWindow(
		"MainWndClass", VERSION, WS_POPUP,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		NULL, NULL, hInst, NULL);
	if (!hwndMain) {
		MessageBox(NULL, "Failed to create window!", "Error", MB_OK);
		return FALSE;
	}

	LoadINIFile(iniFilename);
	originalRotation = rotation;
	if (!initialize()) {	// Load all the mame stuff
		MessageBox(hwndMain, "Initialization failed", "Error", MB_OK);
		return 1;
	}

	// The ini file might have a selected game that is no longer in the list
	// If so, reset to 0
	if (currentGameSelection >= gameDisplayCount)
		currentGameSelection = 0;

	ShowWindow(hwndMain, SW_SHOWMAXIMIZED);
	DrawScreen(hdcOffscreen, hbitmapOffscreen);
	UpdateWindow(hwndMain);
	RemoveCursor();

	retval = runMessageLoop();

	SaveINIFile();
	deinitialize();
	DisplayCursor();

	PlaySound(NULL, NULL, 0);	// Stop any playing sound

	DestroyWindow(hwndMain);

#if DEBUG
	printf("Press enter to close console:\n");
	getchar();
#endif

	return(retval);
}

int runMessageLoop() {
	LARGE_INTEGER start, end, freq;
	LONGLONG update_ticks_pc;
	MSG msg;

	// Get ticks-per-second of the performance counter.
	//   Note the necessary typecast to a LARGE_INTEGER structure
	if (!QueryPerformanceFrequency((LARGE_INTEGER*)&freq)) {
		MessageBox(hwndMain, "Hardware doesn't support QueryPerformanceFrequency()!", "Error", MB_OK);
		return -1;
	}

	// Convert milliseconds per move to performance counter units per move.
	update_ticks_pc = UPDATE_TICKS_MS * freq.QuadPart / 1000;

	// Initialize the counter.
	QueryPerformanceCounter((LARGE_INTEGER*)&start);

	// the main message loop begins here
	while (TRUE) {
		if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			if (!GetMessage(&msg, NULL, 0, 0)) break;
  			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		QueryPerformanceCounter((LARGE_INTEGER*)&end);

		if ((end.QuadPart - start.QuadPart) >= update_ticks_pc) {
			ProcessKeyStrokes(hwndMain);
			start.QuadPart += update_ticks_pc;
		} else {
			Sleep(1);	// This sleeps for 10ms on my machine
		}
	} // End of message loop.

	return msg.wParam;
}

BOOL initialize() {
	char tmpPath[255];
	char tmpFile[255];
	char command[255];
	char message[255];
	int i;

	displayMessage("Loading Game Information From Mame...");

	// Load all the mame stuff here
	if (GetTempPath(255, tmpPath)) {
		// get game list from mame
		if (GetTempFileName(tmpPath, "mam", 0, tmpFile)) {
			sprintf(command, "%s -listinfo", mameExe);
			if (execCommandToOutputFile(command, tmpPath, tmpFile, NULL)) {
				displayMessage("Building Game List...");
				if (!loadListinfoFile(tmpFile)) {
					DeleteFile(tmpFile);
					MessageBox(hwndMain, "Error loading game list!", "Error", MB_OK);
					return FALSE;
				}
				DeleteFile(tmpFile);
			} else {
				sprintf(message, "Error launching [%s]!", command);
				MessageBox(hwndMain, message, "Error", MB_OK);
				DeleteFile(tmpFile);
				return FALSE;
			}
		} else {
			MessageBox(hwndMain, "Error getting temporary file!", "Error", MB_OK);
			return FALSE;
		}

	} else {
		MessageBox(hwndMain, "Error getting temporary path!", "Error", MB_OK);
		return FALSE;
	}

	// If it got here, the gameList is loaded
	displayMessage("Filtering And Sorting Roms...");

	// Determine which games exist
	checkExistance();

	// Sort the full game list (so all short lists will be alpha order)
	qsort(gameList, gameCount, sizeof(GAME *), qsortCompare);

	// Filter games
	buildDisplayGameList();

	// Clear out any stray key presses
	for (i=0; i<255; i++) GetAsyncKeyState(i);

	// Seed the random number generator (PJR 10/2003)
	srand( (unsigned)time( NULL ) );

	// Set the selected game
	firstGameOnScreen = 0;	// this is updated automatically
	currentGameSelection = MAX(findDisplayGameIndex(selectedGameName), 0);
	redrawBrowseMode = TRUE;

	closeMessage();

	inputInit();

	PlayWAVFile(wavStartup, SND_ASYNC);

	if (gameCount == 0) {
		MessageBox(hwndMain, "No games found!", "Error", MB_OK);
		return FALSE;
	} else {
#if DEBUG
		printf("Total/displayed roms: %i, %i\n", gameCount, gameDisplayCount);
#endif
		return TRUE;
	}
}

BOOL deinitialize() {
	freeGameList();

	if (fontFace) free(fontFace);
	if (mameExe) free(mameExe);
	if (mameInfoDat) free(mameInfoDat);
	if (romsDir) free(romsDir);
	if (ssDir) free(ssDir);
	if (flyerDir) free(flyerDir);
	if (wavStartup) free(wavStartup);
	if (wavShutdown) free(wavShutdown);
	if (wavStartGame) free(wavStartGame);
	if (wavError) free(wavError);
	if (statusBarFontFace) free(statusBarFontFace);
	if (selectedGameName) free(selectedGameName);

	if (hbitmapOffscreen) DeleteObject(hbitmapOffscreen);
	if (hdcOffscreen) DeleteDC(hdcOffscreen);
	if (hbitmapTile) DeleteObject(hbitmapTile);
	if (hdcTile) DeleteDC(hdcTile);
	ReleaseDC(hwndMain, hdcMain);

	return TRUE;
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch(uMsg) {
		case WM_CREATE:
			// get the main dc
			hdcMain = GetDC(hwnd);
			hdcOffscreen = CreateCompatibleDC(hdcMain);
			hdcTile = CreateCompatibleDC(hdcMain);
			break;
		case WM_SIZE:
			{
				RECT rect;
				long width;
				long height;
				HBITMAP hOldBitmap;

				// get the width & height of the window
				GetClientRect(hwnd, &rect);
				width = rect.right;
				height = rect.bottom;

				// create the offscreen dc & bitmap
				hbitmapOffscreen = CreateCompatibleBitmap(hdcMain, width, height);
				hOldBitmap = (HBITMAP)SelectObject(hdcOffscreen, hbitmapOffscreen);
				DeleteObject(hOldBitmap);

				// create the tile dc & bitmap
				hbitmapTile = CreateCompatibleBitmap(hdcMain, width, height);
				hOldBitmap = (HBITMAP)SelectObject(hdcTile, hbitmapTile);
				DeleteObject(hOldBitmap);

				break;
			}
		case WM_DESTROY:
			PostQuitMessage(0);
			break;
		case WM_PAINT:
			{
				PAINTSTRUCT paintStruct;
				RECT rect;
				BeginPaint(hwnd, &paintStruct);

				GetClientRect(hwnd, &rect);

				if(hdcOffscreen)
					BitBlt(hdcMain, 0, 0, rect.right, rect.bottom, hdcOffscreen, 0, 0, SRCCOPY);

				EndPaint(hwnd, &paintStruct);
				break;
			}
		case WM_SETFOCUS:
			inputEnabled = TRUE;
			break;
		case WM_KILLFOCUS:
			inputEnabled = FALSE;
			break;
		default:
			return(DefWindowProc(hwnd, uMsg, wParam, lParam));
	}
	return 0;
}

void RemoveCursor() {
	while (ShowCursor(FALSE) >= 0);
}

void DisplayCursor() {
	while (ShowCursor(TRUE) < 0) ;
}

// sync should be SND_SYNC or SND_ASYNC
void PlayWAVFile(char *filename, DWORD sync) {
	if (filename) PlaySound(filename, NULL, sync | SND_NODEFAULT | SND_FILENAME);
}

void ProcessKeyStrokes(HWND hwnd) {
	int i, newRotation;
	char message[255];
	DWORD exitCode;
	BOOL p1Up, p1Down, p1Pgup, p1Pgdown, p1Exit, p1Start, p1Info, p1Menu;
	BOOL p2Up, p2Down, p2Pgup, p2Pgdown, p2Exit, p2Start, p2Info, p2Menu;

	// Record the states of all the keys
	p1Up = keyPressed(keySelectUp) || joyInPosition(joySelectUp);
	p1Down = keyPressed(keySelectDown) || joyInPosition(joySelectDown);
	p1Pgup = keyPressed(keySelectPageDown) || joyInPosition(joySelectPageDown);
	p1Pgdown = keyPressed(keySelectPageUp) || joyInPosition(joySelectPageUp);
	p1Exit = keyPressed(keyExit) || joyInPosition(joyExit);
	p1Start = keyPressed(keyStartGame) || joyInPosition(joyStartGame);
	p1Info = keyPressed(keyGameInfo) || joyInPosition(joyGameInfo);
	p1Menu = keyPressed(keyShowMenu) || joyInPosition(joyShowMenu);

	p2Up = keyPressed(keyP2SelectUp);
	p2Down = keyPressed(keyP2SelectDown);
	p2Pgup = keyPressed(keyP2SelectPageDown);
	p2Pgdown = keyPressed(keyP2SelectPageUp);
	p2Exit = keyPressed(keyP2Exit);
	p2Start = keyPressed(keyP2StartGame);
	p2Info = keyPressed(keyP2GameInfo);
	p2Menu = keyPressed(keyP2ShowMenu);

	// Flip the screen if requested
	if (invertPlayer2Screen) {
		newRotation = rotation;
		switch (originalRotation) {
			case ROTATION_NORMAL:
				if (p2Up ||  p2Down || p2Pgup || p2Pgdown || p2Exit || p2Start || p2Info || p2Menu)
					newRotation = ROTATION_INVERTED;
				if (p1Up || p1Down || p1Pgup || p1Pgdown || p1Exit || p1Start || p1Info || p1Menu)
					newRotation = ROTATION_NORMAL;
				break;
			case ROTATION_INVERTED:
				if (p2Up ||  p2Down || p2Pgup || p2Pgdown || p2Exit || p2Start || p2Info || p2Menu)
					newRotation = ROTATION_NORMAL;
				if (p1Up || p1Down || p1Pgup || p1Pgdown || p1Exit || p1Start || p1Info || p1Menu)
					newRotation = ROTATION_INVERTED;
				break;
			case ROTATION_ROR:
				if (p2Up ||  p2Down || p2Pgup || p2Pgdown || p2Exit || p2Start || p2Info || p2Menu)
					newRotation = ROTATION_ROL;
				if (p1Up || p1Down || p1Pgup || p1Pgdown || p1Exit || p1Start || p1Info || p1Menu)
					newRotation = ROTATION_ROR;
				break;
			case ROTATION_ROL:
				if (p2Up ||  p2Down || p2Pgup || p2Pgdown || p2Exit || p2Start || p2Info || p2Menu)
					newRotation = ROTATION_ROR;
				if (p1Up || p1Down || p1Pgup || p1Pgdown || p1Exit || p1Start || p1Info || p1Menu)
					newRotation = ROTATION_ROL;
				break;
		}
		if (rotation != newRotation) {
			rotation = newRotation;
			redrawBrowseMode = TRUE;
		}
	}

	if (p1Up || p2Up) {		// Select next lower # game
		if (currentGameSelection > 0) {
			currentGameSelection--;
			DrawScreen(hdcOffscreen, hbitmapOffscreen);
			InvalidateRect(hwnd, NULL, FALSE);
			PlayWAVFile(wavSelect, SND_ASYNC);
		} else
			PlayWAVFile(wavError, SND_ASYNC);
	}

	if (p1Down || p2Down) {	// Select next higher # game
		if (currentGameSelection < gameDisplayCount-1) {
			currentGameSelection++;
			DrawScreen(hdcOffscreen, hbitmapOffscreen);
			InvalidateRect(hwnd, NULL, FALSE);
			PlayWAVFile(wavSelect, SND_ASYNC);
		} else
			PlayWAVFile(wavError, SND_ASYNC);
	}

	if (p1Pgdown || p2Pgdown) {	// Select One Page up
		RECT rect;
		GetClientRect(hwnd, &rect);
		if (currentGameSelection > 0) {
			if(currentGameSelection < 0) currentGameSelection = 0;
 			if (menuType == MENUTYPE_LISTVIEW)
				currentGameSelection = MAX(currentGameSelection - GetNumDisplayRows(hdcOffscreen, rect.right, rect.bottom*2, statusBarHeight), 0);
			else
				currentGameSelection = MAX(currentGameSelection -
					GetNumImagesY(hdcOffscreen, rect.right, rect.bottom, statusBarHeight), 0);

			DrawScreen(hdcOffscreen, hbitmapOffscreen);
			InvalidateRect(hwnd, NULL, FALSE);
			UpdateWindow(hwnd);
			PlayWAVFile(wavSelect, SND_ASYNC);
		} else
			PlayWAVFile(wavError, SND_ASYNC);
	}

	if(p1Pgup || p2Pgup) { // Select one page down
		RECT rect;
		GetClientRect(hwnd, &rect);
		if (currentGameSelection < gameDisplayCount-1) {
			if(currentGameSelection > gameDisplayCount) currentGameSelection = gameDisplayCount;
			if (menuType == MENUTYPE_LISTVIEW)
				currentGameSelection = MIN(currentGameSelection + GetNumDisplayRows(hdcOffscreen, rect.right, rect.bottom*2, statusBarHeight), gameDisplayCount-1);
			else
				currentGameSelection = MIN(currentGameSelection +
					GetNumImagesY(hdcTile, rect.right, rect.bottom, statusBarHeight), gameDisplayCount-1);

			DrawScreen(hdcOffscreen, hbitmapOffscreen);
			InvalidateRect(hwnd, NULL, FALSE);
			UpdateWindow(hwnd);
			PlayWAVFile(wavSelect, SND_ASYNC);
		} else
			PlayWAVFile(wavError, SND_ASYNC);
	}

	if (keyPressed(0x20)) {	// Select 100 games up
		RECT rect;
		GetClientRect(hwnd, &rect);
		if (currentGameSelection > 0) {
			currentGameSelection -=100;
			if(currentGameSelection < 0) currentGameSelection = 0;
			DrawScreen(hdcOffscreen, hbitmapOffscreen);
			InvalidateRect(hwnd, NULL, FALSE);
			UpdateWindow(hwnd);
			PlayWAVFile(wavSelect, SND_ASYNC);
		} else
		PlayWAVFile(wavError, SND_ASYNC);
	}

	if (keyPressed(0x58)) {// Select 100 games down
		RECT rect;
		GetClientRect(hwnd, &rect);
		if (currentGameSelection < gameDisplayCount-1) {
			currentGameSelection +=100;
			if(currentGameSelection > gameDisplayCount) currentGameSelection = gameDisplayCount - 1;
			DrawScreen(hdcOffscreen, hbitmapOffscreen);
			InvalidateRect(hwnd, NULL, FALSE);
			UpdateWindow(hwnd);
			PlayWAVFile(wavSelect, SND_ASYNC);
		} else
			PlayWAVFile(wavError, SND_ASYNC);
	}

	if (keyPressed(0x12)) {	// ALT (menu) key-  Random game
		RECT rect;
		GetClientRect(hwnd, &rect);
		currentGameSelection = gameDisplayCount * rand() / (RAND_MAX + 1.0);
		DrawScreen(hdcOffscreen, hbitmapOffscreen);
		InvalidateRect(hwnd, NULL, FALSE);
		UpdateWindow(hwnd);
		PlayWAVFile(wavSelect, SND_ASYNC);
	}

	if (keyPressed(0x5a)) {	// Z key- Change sort order
		RECT rect;
		GetClientRect(hwnd, &rect);
		currentGameSelection = gameDisplayCount * rand() / (RAND_MAX + 1.0);
		sortBy++; if(sortBy>2) sortBy=0;
		displayMessage("Filtering And Sorting Roms...");
		qsort(gameList, gameCount, sizeof(GAME *), qsortCompare);
		buildDisplayGameList();
		closeMessage();
		DrawScreen(hdcOffscreen, hbitmapOffscreen);
		InvalidateRect(hwnd, NULL, FALSE);
		UpdateWindow(hwnd);
		PlayWAVFile(wavSelect, SND_ASYNC);
	}


	if (keyPressed(0x10)) {	// SHIFT key-  Change previewPicture
		RECT rect;
		GetClientRect(hwnd, &rect);
		previewPictures++; if(previewPictures>3) previewPictures=0;
		DrawScreen(hdcOffscreen, hbitmapOffscreen);
		InvalidateRect(hwnd, NULL, FALSE);
		UpdateWindow(hwnd);
		PlayWAVFile(wavSelect, SND_ASYNC);
	}

	if (p1Exit || p2Exit) {			// Exit the system
		PlayWAVFile(wavShutdown, SND_SYNC);
		PostQuitMessage(0);
	}

	if (p1Start || p2Start) {		// Start the selected game
		if(!gameDisplayList[currentGameSelection]->exists == TRUE) {
			sprintf(message, "Sorry, you appear to be missing these ROMs...");
			displayMessage(message);
			Sleep(1500);
			closeMessage();
		}
		else {
			if(attractAVI) MCIWndPause(attractAVI);
			sprintf(message, "Launching Game: %s...", gameDisplayList[currentGameSelection]->description);
			displayMessage(message);
			PlayWAVFile(wavStartGame, SND_SYNC);
			launchMame(gameDisplayList[currentGameSelection], &exitCode);
			if (exitCode != 0) {
				// Throw up a message if mame barfs
				PlayWAVFile(wavError, SND_ASYNC);
				displayMessage("Error Launching Mame!");
				Sleep(3000);
			}
			// Pause a moment to avoid escape registering twice
			Sleep(500);
			// Clear out any stray key presses coming back from mame
			for (i=0; i<255; i++) GetAsyncKeyState(i);
			closeMessage();
			MCIWndResume(attractAVI);
			InvalidateRect(hwnd, NULL, FALSE);
		}
	}

	if (p1Info || p2Info) {		// Show info about the selected game
		if (useMameInfoDat) {
			char **lines;
			int lineCount, parentIndex;

			lines=NULL;
			lineCount=0;

			if (gameDisplayList[currentGameSelection]->numMameInfoLines > 0)  {       // sometimes numlines is 4xxxxxxx
				//(gameDisplayList[currentGameSelection]->numMameInfoLines < 500)) {   // there is a bug and this is a hack PJR
				lines = gameDisplayList[currentGameSelection]->mameInfoLines;
				lineCount = gameDisplayList[currentGameSelection]->numMameInfoLines;
			} else if (gameDisplayList[currentGameSelection]->cloneof) {
				parentIndex = findDisplayGameIndex(gameDisplayList[currentGameSelection]->cloneof);
				if (parentIndex > -1) {
					lines = gameDisplayList[parentIndex]->mameInfoLines;
					lineCount = gameDisplayList[parentIndex]->numMameInfoLines;
				}
			}

			if (lineCount > 0) {
				displayScrollingTextScreen(lines, lineCount);

				// Pause a moment to avoid escape registering twice
				Sleep(300);
				// Clear out any stray key presses coming back from mame
				for (i=0; i<255; i++) GetAsyncKeyState(i);
			}
			else {
				sprintf(message,"No game information found.");
				displayMessage(message);
				Sleep(1500);
				closeMessage();
			}
		}
	}

	if (p1Menu || p2Menu) {		// Show the config menu
		if (runConfigMenu()) {
			char selectedGameName[255];
			Sleep(200);	// Allow the menu time to disappear

			// Store the name of the selected game
			if (gameDisplayCount > 0)
				strcpy(selectedGameName, gameDisplayList[currentGameSelection]->romname);

			// Re-filter games
			buildDisplayGameList();

			// Display the right game
			if (gameDisplayCount > 0)
				currentGameSelection = findDisplayGameIndex(selectedGameName);
			else
				currentGameSelection = 0;

			// Redraw the screen
			redrawBrowseMode = TRUE;
			DrawScreen(hdcOffscreen, hbitmapOffscreen);
			InvalidateRect(hwnd, NULL, FALSE);
		}
		// Pause a moment to avoid escape registering twice
		Sleep(300);
		// Clear out any stray key presses coming back from mame
		for (i=0; i<255; i++) GetAsyncKeyState(i);
	}
}

